/****************************************************************************
Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.
****************************************************************************/

#ifndef BBL_DICOM_H
#define BBL_DICOM_H

#ifndef NULL
#define NULL 0
#endif

#include <stdio.h>
#include <sys/stat.h>

/*  FIXME: SunOS 4.1.3 does not have this constant in the same place as
 *  Solaris 2.x; I am too lazy to go find it.  -DRW(?) */
#ifndef FILENAME_MAX
#define FILENAME_MAX 256
#endif

/*
 * Definitions that allow this header file to be used either with or
 * without ANSI C features like function prototypes.  Taken from tcl.h.
 */

#undef _ANSI_ARGS_
#undef CONST
#if ((defined(__STDC__) || defined(SABER)) && !defined(NO_PROTOTYPE)) || defined(__cplusplus)
#   define _USING_PROTOTYPES_ 1
#   define _ANSI_ARGS_(x)	x
#   define CONST const
#   ifdef __cplusplus
#       define VARARGS(first) (first, ...)
#   else
#       define VARARGS(first) ()
#   endif
#else
#   define _ANSI_ARGS_(x)	()
#   define CONST
#endif


/*
 *  Definition storage limits
 */
#define MAX_ELEMENT_NAME_LEN 32
#define MAX_VALREP_LEN 2
#define MAX_VALMULT_LEN 5

#define INVALID_MESSAGE		-1
#define ELEMENT_NOTFOUND	-2

#define MSG_OPEN		0x0001
#	define MSG_CLOSED	0x0000
#	define MSG_OPENED	0x0001
#define MSG_MODE		0x0006
#	define MSG_READ		0x0002
#	define MSG_WRITE	0x0004
#	define MSG_APPEND	0x0006

/*
 *  Maximum length of any string element stored in memory.
 *    This excludes the following Value Representations:
 *
 *    ELEM_VR_OB, ELEM_VR_OW, (Other Byte String, Other Word String),
 *       neither of which have a maximum length.
 *
 *    ELEM_VR_LT, ELEM_VR_ST, (Long Text, Short Text),
 *       which have a maximum length of 10240 and 1024, respectively
 */
#define MAX_STRING_ELEM_VALUE_LEN	64
#define MAX_TEXT_ELEM_VALUE_LEN		10240

/*
 *  Element Value Representation
 *
 *	Code:	Meaning and interpretation:
 *
 *	(Stored in memory, strings)
 *
 *	AE	Application Entity: max. 16 bytes (any, except LF, FF, CR)
 *	AS	Age String: max. 4 bytes (dddD, dddW, dddM, or dddY)
 *	CS	Code String: max. 16 bytes (digits, uppercase, space, "_")
 *	DA	Date: max. 8 or 10 bytes (new: yyyymmdd, old: yyyy.mm.dd)
 *	DS	Decimal String: max. 16 bytes (digits, ".", "e", "+", "-")
 *	DT	Date Time: max. 26 bytes (yyyymmddhhmmss.ffffff&zzzz)
 *	IS	Integer String: max. 12 bytes (digits, "+", "-")
 *	LO	Long String: max. 64 chars (any)
 *	PN	Person Name: max. 64 chars (any, except LF, FF, CR)
 *	SH	Short String: max. 16 chars (any)
 *	TM	Time: max. 16 bytes (hhmmss.ffffff), 24-hour clock
 *
 *      (Stored in memory, binary)
 *
 *	UL	Unsigned Long: 32-bit binary integer, 0 < n < 2^32
 *	US	Unsigned Short: 16-bit binary integer, 0 < n < 2^16
 *
 *	(Kept in file until accessed)
 *
 *	LT	Long Text: max. 10240 chars (any)
 *	OB	Other Byte String: 8-bit words; Little/Big-Endian sensitive
 *	OW	Other Word String: 16-bit words; Little/Big-Endian sensitive
 *	ST	Short Text: max. 1024 chars (any)
 *
 *  Hash table:
 *
 *    hash code = first character * 256 + second character
 *
 *    (will only work if internal character representation is
 *    equal to stored character representation, i.e. both are
 *    one-byte unsigned chars, or two-byte, or wide-chars, etc.).
 */


#define ELEM_VR_AE	('A' * 256 + 'E')
#define ELEM_VR_AS	('A' * 256 + 'S')
#define ELEM_VR_CS	('C' * 256 + 'S')
#define ELEM_VR_DA	('D' * 256 + 'A')
#define ELEM_VR_DS	('D' * 256 + 'S')
#define ELEM_VR_DT	('D' * 256 + 'T')
#define ELEM_VR_IS	('I' * 256 + 'S')
#define ELEM_VR_LO	('L' * 256 + 'O')
#define ELEM_VR_PN	('P' * 256 + 'N')
#define ELEM_VR_SH	('S' * 256 + 'H')
#define ELEM_VR_TM	('T' * 256 + 'M')

#define ELEM_VR_UL	('U' * 256 + 'L')
#define ELEM_VR_US	('U' * 256 + 'S')

#define ELEM_VR_LT	('L' * 256 + 'T')
#define ELEM_VR_OB	('O' * 256 + 'B')
#define ELEM_VR_OW	('O' * 256 + 'W')
#define ELEM_VR_ST	('S' * 256 + 'T')

/*
 *  Element Value Multiplicity
 *
 *	"x"	x values, x is an integer
 *	"1-n"	any number of values possible, delimited
 */
#define ELEM_VM_1_THRU_N	-1
#define ELEM_VM_UNDEF		-2

/*
 *  Element Value Location (for type specifier)
 */
#define ELEM_VL_UNDEF		0
#define ELEM_VL_MEM		1
#define ELEM_VL_FILE		2

/* FIXME:  The use of BIG_ENDIAN and LITTLE_ENDIAN here conflicts with the
   same names used in the Gnu header file endian.h, both as to meaning and
   coding.  It appears that bbl_dicom.c gets the right definitions for its
   purposes, but beware of future problems.  -PWH 19981216 */
/*
 *  Transfer syntax options:
 *  ACRNEMA is in Big-Endian format,
 *  DICOM is in Little-Endian format (default for DICOM v3.0).
 */
#define BIG_ENDIAN 0
#define LITTLE_ENDIAN 1



typedef unsigned short	ElementGroupId;
typedef unsigned short	ElementId;
typedef unsigned long	ElementLength;
typedef off_t		ElementValueOffset;
typedef void		*ElementValue;



/*********************************************************************
 *  Element Definition typedefs
 *********************************************************************/

/*
 *  Element ID, Description,
 *  hashed Element Data Type,
 *  converted Element Multiplicity
 */
typedef struct definition * DefinitionPtr;
typedef struct definition {
  ElementGroupId	elementGroupId;
  ElementId		elementId;
  char			elementName[MAX_ELEMENT_NAME_LEN + 1];
  int			valueRepresentation;
  int			valueMultiplicity;
} Definition;
#define VALREP( Definition ) ((Definition)->valueRepresentation)
#define VALMULT( Definition ) ((Definition)->valueMultiplicity)

#define EgroupLength			0x0000

#define GstudyInfo			0x0008

#define EstudyDate			0x0020
#define EstudyTime			0x0030
#define Emodality			0x0060
#define EinstitutionName		0x0080
#define EreferringPhysiciansName	0x0090


#define GpatientInfo			0x0010

#define Epatient_name			0x0010
#define Epatient_id			0x0020
#define Epatient_birthDate		0x0030
#define Epatient_sex			0x0040
#define Epatient_age			0x1010
#define Epatient_weight			0x1030


#define GphysicalInfo			0x0018

#define Emr_acq_type			0x0023
#define Esequence_name			0x0024
#define Eslice_thickness		0x0050
#define Etr				0x0080
#define Ete				0x0081
#define Eecho_num			0x0086
#define EsliceSpacing			0x0088
#define Eserial_num			0x1000


#define GstorageInfo			0x0020

#define Estudy_id			0x0010
#define EseriesNumber			0x0011
#define Eslice_num			0x0013
#define Eorientation			0x0035
#define Estudy_series			0x1000
#define Etotalimages			0x1002
#define EsliceLocation			0x1041
#define Ecomment			0x4000


#define GimagingInfo			0x0028

#define Enum_dims			0x0005
#define Erows				0x0010
#define Ecolumns			0x0011
#define EpixelSpacing			0x0030
#define EbitsAllocated			0x0100
#define EbitsStored			0x0101
#define EhighBit			0x0102


#define GpixelData			0x7fe0

#define Edata				0x0010

/*********************************************************************
 *  Element Value typedefs
 *********************************************************************/

typedef struct element * ElementPtr;
typedef struct element {
  ElementGroupId	groupId;
  ElementId		id;
  ElementLength		length;
  int			multiplicity;
  ElementPtr		nextElement;
  int			valueLocation;
  size_t		valueSize;
  union {

    /*
     *  In-memory storage of small strings and binary values.
     *    malloc()'ed when Element is created.
     */
    ElementValue	ptr;

    /*
     *  If value size is large, we need to keep track of the value's
     *    offset into the message to save search time when we actually
     *    need to access the data.
     */
    ElementValueOffset	offset;

  }			value;
} Element;

typedef struct elementGroup * ElementGroupPtr;
typedef struct elementGroup {
  ElementGroupId	groupId;
  ElementPtr		firstElement;
  ElementPtr		lastElement;
  ElementGroupPtr	nextGroup;
} ElementGroup;

typedef struct message * MessagePtr;
typedef struct message {
  int			error;
  char			fileName[FILENAME_MAX];
  FILE			*fid;
  int			status;
  off_t			size;
  int			transferSyntax;
  unsigned long		numElementGroups;
  unsigned long		numElements;
  ElementGroupPtr	firstGroup;
  ElementGroupPtr	lastGroup;
} Message;




#define swapUShort( uShortValue ) \
	((unsigned short) \
	 (((unsigned short) (uShortValue & 0x00FF) << 8) | \
	  ((unsigned short) (uShortValue & 0xFF00) >> 8)))
#define swapULong( uLongValue ) \
	((unsigned long) \
	 (((uLongValue & 0x000000FF) << 24) | \
	  ((uLongValue & 0x0000FF00) <<  8) | \
	  ((uLongValue & 0x00FF0000) >>  8) | \
	  ((uLongValue & 0xFF000000) >> 24)))

/*
 *  Function prototypes
 */
DefinitionPtr findElementDefn _ANSI_ARGS_(( ElementGroupId, ElementId ));

MessagePtr msgOpen _ANSI_ARGS_(( char *, int ));
void msgClose _ANSI_ARGS_(( MessagePtr ));
CONST ElementValue getValue _ANSI_ARGS_(( MessagePtr, ElementPtr ));
CONST char * elementToString _ANSI_ARGS_(( MessagePtr, ElementPtr ));
void initMessage _ANSI_ARGS_(( MessagePtr ));
ElementPtr addElement _ANSI_ARGS_(( MessagePtr, ElementGroupId, ElementId,
				   ElementLength, ElementValueOffset ));
ElementPtr dupElement _ANSI_ARGS_(( ElementPtr ));
void freeElement _ANSI_ARGS_(( ElementPtr ));
ElementPtr findElement _ANSI_ARGS_(( MessagePtr, ElementGroupId, ElementId ));
int getElements _ANSI_ARGS_(( MessagePtr ));
void dumpElement _ANSI_ARGS_(( MessagePtr, ElementGroupId, ElementId ));
void dumpElements _ANSI_ARGS_(( MessagePtr ));

#endif
